<?php
/**
 * This file is part of PHPWord - A pure PHP library for reading and writing
 * word processing documents.
 *
 * PHPWord is free software distributed under the terms of the GNU Lesser
 * General Public License version 3 as published by the Free Software Foundation.
 *
 * For the full copyright and license information, please read the LICENSE
 * file that was distributed with this source code. For the full list of
 * contributors, visit https://github.com/PHPOffice/PHPWord/contributors.
 *
 * @link        https://github.com/PHPOffice/PHPWord
 * @copyright   2010-2014 PHPWord contributors
 * @license     http://www.gnu.org/licenses/lgpl.txt LGPL version 3
 */

namespace PhpOffice\PhpWord\Reader\ODText;

use PhpOffice\PhpWord\Reader\Word2007\AbstractPart as Word2007AbstractPart;
use PhpOffice\PhpWord\Shared\XMLReader;
use PhpOffice\PhpWord\Reader\ODText\Style;
use PhpOffice\PhpWord\Style\Font;
use PhpOffice\PhpWord\Style\Paragraph;

/**
 * Abstract part reader
 *
 * @since 0.10.0
 * @codeCoverageIgnore
 */
abstract class AbstractPart extends Word2007AbstractPart
{
	const STYLE_TYPE_PARAGRAPH = 'paragraph';
	const STYLE_TYPE_FONT = 'font';
	const STYLE_TYPE_COLUMN = 'column';
	const STYLE_TYPE_ROW = 'row';
	const STYLE_TYPE_CELL = 'cell';
	const STYLE_TYPE_LIST = 'list';

	const READ_PARSER = 'attributeParse';

	const INDENT_BASE_VALUE_ODT = 1.251;
	const INDENT_BASE_VALUE_DOCX = 240;

	/**
	 * Read style:paragraph-properties
	 *
	 * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
	 * @param \DOMElement $domNode
	 * @return array|null
	 */
	protected function readParagraphStyle(XMLReader $xmlReader, \DOMElement $styleNode, $parentStyleName = null)
	{
		$styleDefs = array(
			'indent'           => array(self::READ_PARSER, 'fo:margin-left', 'parseMargin'),
			'hanging'          => array(self::READ_PARSER, 'fo:text-indent', 'parseIndent'),
			'align'            => array(self::READ_PARSER, 'fo:text-align', 'parseAlign'),
			'borderBottomSize' => array(self::READ_PARSER, 'fo:border-bottom', 'parseBorderBottom')
		);

		return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs, $parentStyleName);
	}

	/**
	 * Read style:text-properties
	 *
	 * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
	 * @param \DOMElement $domNode
	 * @return array|null
	 */
	protected function readFontStyle(XMLReader $xmlReader, \DOMElement $styleNode, $parentStyleName = null)
	{
		$styleDefs = array(
			'name'          => array(self::READ_VALUE, 'style:font-name'),
			'size'          => array(self::READ_PARSER, 'fo:font-size', 'parseFontSize'),
			'underline'     => array(self::READ_PARSER, 'style:text-underline-style', 'parseUnderline'),
			'bold'          => array(self::READ_PARSER, 'fo:font-weight', 'parseWeight'),
			'italic'        => array(self::READ_PARSER, 'fo:font-style', 'parseStyle'),
			'strikethrough' => array(self::READ_EQUAL, 'style:text-line-through-style', 'solid'),
			'superScript'   => array(self::READ_EQUAL, 'style:text-position', 'super 58%'),
			'subScript'     => array(self::READ_EQUAL, 'style:text-position', 'sub 58%'),
			'color'         => array(self::READ_PARSER, 'fo:color', 'parseColor'),
			'bgColor'       => array(self::READ_PARSER, 'fo:background-color', 'parseColor')
		);

		return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs, $parentStyleName);
	}

	protected function readColumnStyle(XMLReader $xmlReader, \DOMElement $styleNode)
	{
		$styleDefs = array(
			'width' => array(self::READ_PARSER, 'style:column-width', 'parseColumnWidth')
		);

		return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs);
	}

	protected function readRowStyle(XMLReader $xmlReader, \DOMElement $styleNode)
	{
		$styleDefs = array(
			'height'      => array(self::READ_PARSER, 'style:min-row-height', 'parseRowHeight'),
			'exactHeight' => array(self::READ_TRUE, 'style:min-row-height')
		);

		return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs);
	}

	protected function readCellStyle(XMLReader $xmlReader, \DOMElement $styleNode)
	{
		$styleDefs = array(
			'valign' => array(self::READ_TRUE, 'style:vertical-align')
		);
		$styles = $this->readStyleDefs($xmlReader, $styleNode, $styleDefs);

		$this->readBorderStyle($xmlReader, $styleNode, $styles, 'fo:border', '');
		$this->readBorderStyle($xmlReader, $styleNode, $styles, 'fo:border-left', 'Left');
		$this->readBorderStyle($xmlReader, $styleNode, $styles, 'fo:border-right', 'Right');
		$this->readBorderStyle($xmlReader, $styleNode, $styles, 'fo:border-top', 'Top');
		$this->readBorderStyle($xmlReader, $styleNode, $styles, 'fo:border-bottom', 'Bottom');

		return $styles;
	}

	private function readBorderStyle(XMLReader $xmlReader, \DOMElement $styleNode, &$styles, $attrName, $position)
	{
		$value = $xmlReader->getAttribute($attrName, $styleNode);

		if (empty($value)) {
			return;
		}

		$color = 'aaaaaa';
		$style = 'single';
		$size = 8;

		if ('none' === $value) {
			$color = '';
			$style = 'nil';
			$size = '';
		} else {
			$borderStyle = $this->parseBorderStyle($value);
			list($size, $style, $color) = $borderStyle;
		}

		if ('' === $position) {
			$styles['borderLeftColor'] = $color;
			$styles['borderLeftStyle'] = $style;
			$styles['borderLeftSize'] = $size;
			$styles['borderTopColor'] = $color;
			$styles['borderTopStyle'] = $style;
			$styles['borderTopSize'] = $size;
			$styles['borderRightColor'] = $color;
			$styles['borderRightStyle'] = $style;
			$styles['borderRightSize'] = $size;
			$styles['borderBottomColor'] = $color;
			$styles['borderBottomStyle'] = $style;
			$styles['borderBottomSize'] = $size;
		} else {
			$styles['border'.$position.'Color'] = $color;
			$styles['border'.$position.'Style'] = $style;
			$styles['border'.$position.'Size'] = $size;
		}
	}

	/**
     * Read style definition
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $parentNode
     * @param array $styleDefs
     * @ignoreScrutinizerPatch
     * @return array
     */
	protected function readStyleDefs(XMLReader $xmlReader, \DOMElement $domNode = null, $styleDefs = array(), $parentStyleName = null)
	{
		$styles = array();

		foreach ($styleDefs as $styleProp => $styleVal) {
			@list($method, $attribute, $expected) = $styleVal;

			$attributeValue = $xmlReader->getAttribute($attribute, $domNode);

			if (empty($attributeValue)) {
				continue;
			}

			$styleValue = $this->readStyleDef($method, $attributeValue, $expected, $parentStyleName);

			if ($styleValue !== null) {
				$styles[$styleProp] = $styleValue;
			}
		}

		return $styles;
	}

	/**
     * Return style definition based on conversion method
     *
     * @param string $method
     * @ignoreScrutinizerPatch
     * @param mixed $attributeValue
     * @param mixed $expected
     * @return mixed
     */
	private function readStyleDef($method, $attributeValue, $expected, $parentStyleName)
	{
		$style = $attributeValue;

		if (self::READ_PARSER == $method) {
			$style = $this->$expected($style, $parentStyleName);
		} elseif (self::READ_EQUAL == $method) {
            $style = $attributeValue == $expected;
		} elseif (self::READ_TRUE == $method) {
			$style = true;
		}

		return $style;
	}

	private function parseDecimal($value)
	{
		if (preg_match('/^-*\d+\.*\d*/', $value, $matches)) {
			return floatval($matches[0]);
		}
		return 0;
	}

	private function parseMargin($value)
	{
		if (0 < ($margin = $this->parseDecimal($value))) {
			$multiplier = $margin / self::INDENT_BASE_VALUE_ODT;
			return $multiplier * self::INDENT_BASE_VALUE_DOCX;
		}
		return 0;
	}

	private function parseIndent($value)
	{
		$indent = abs($this->parseDecimal($value));
		$multiplier = $indent / self::INDENT_BASE_VALUE_ODT;
		return $multiplier * self::INDENT_BASE_VALUE_DOCX;
	}

	private function parseAlign($value)
	{
		switch ($value) {
			case 'start':
				return 'left';
			case 'end':
				return 'right';
			case 'justify':
				return 'both';
			default:
				return $value;
		}
	}

	private function parseBorderBottom($value)
	{
		if (0 < strlen($value)) {
			return "6"; //magic number
		}
	}

	private function parseFontSize($value, $parentStyleName)
	{
		if (!preg_match('/^\d+/', $value, $matches)) {
			return;
		}

		if (false != strpos($value, '%')) {
			while (true) {
				$parentStyle = Style::$styles[$parentStyleName][self::STYLE_TYPE_FONT];
				if (isset($parentStyle)) {
					$size = $parentStyle['size'];
				}
				$parentStyleName = Style::$styles[$parentStyleName]['parent'];
				if (!isset($parentStyleName)) {
					break;
				}
			}
			if (isset($size)) {
				return floatval($size) * floatval($matches[0]) / 100.0;
			}
		} else {
			return $matches[0];
		}
	}

	private function parseUnderline($value)
	{
		if ('solid' === $value) {
			return 'single';
		}
		return '';
	}

	private function parseWeight($value)
	{
		return ('bold' === $value) ? true : false;
	}

	private function parseStyle($value)
	{
		return ('italic' === $value) ? true : false;
	}

	private function parseColor($value)
	{
		if (7 > strlen($value)) {
			return '';
		}
		return substr($value, 1, 6);
	}

	protected function parseCm($value)
	{
		if (preg_match('/^\d+\.*\d*/', $value, $matches)) {
			return floatval($matches[0]);
		}
	}

	private function parseColumnWidth($value)
	{
		if (preg_match('/^\d+\.*\d*/', $value, $matches)) {
			return round($this->convertCmToDxa($matches[0]));
		}
	}

	private function parseRowHeight($value)
	{
		if (preg_match('/^\d+\.*\d*/', $value, $matches)) {
			return round($this->convertCmToDxa($matches[0]));
		}
	}

	private function convertCmToDxa($valueCm)
	{
		$valueCm = floatval($valueCm);
		return $valueCm / 2.54 * 1440;//cm to inch to dxa
	}

	private function parseBorderStyle($value)
	{
		$borderStyle = array();
		$values = explode(' ', $value);

		if (3 !== count($values)) {
			return;
		}

		if (8 > ($size = $this->parseBorderWidth($values[0]))) {
			$size = 8;
		}
		$color = $this->parseColor($values[2]);

		$borderStyle[] = $size;
		$borderStyle[] = 'single';
		$borderStyle[] = $color;

		return $borderStyle;
	}

	private function parseBorderWidth($value)
	{
		if (preg_match('/^\d+\.*\d*/', $value, $matches)) {
			return \PhpOffice\PhpWord\Shared\Converter::cmToPoint(floatval($matches[0])) * 8;
		}
	}
}
