<?php
/**
 * This file is part of PHPWord - A pure PHP library for reading and writing
 * word processing documents.
 *
 * PHPWord is free software distributed under the terms of the GNU Lesser
 * General Public License version 3 as published by the Free Software Foundation.
 *
 * For the full copyright and license information, please read the LICENSE
 * file that was distributed with this source code. For the full list of
 * contributors, visit https://github.com/PHPOffice/PHPWord/contributors.
 *
 * @link        https://github.com/PHPOffice/PHPWord
 * @copyright   2010-2014 PHPWord contributors
 * @license     http://www.gnu.org/licenses/lgpl.txt LGPL version 3
 */

namespace PhpOffice\PhpWord\Reader\RTF;

use PhpOffice\PhpWord\PhpWord;

/**
 * RTF document reader
 *
 * References:
 * - How to Write an RTF Reader http://latex2rtf.sourceforge.net/rtfspec_45.html
 * - PHP rtfclass by Markus Fischer https://github.com/mfn/rtfclass
 * - JavaScript RTF-parser by LazyGyu https://github.com/lazygyu/RTF-parser
 *
 * @since 0.11.0
 * @SuppressWarnings(PHPMD.UnusedPrivateMethod)
 */
class Document
{
    /** @const int */
    const PARA = 'readParagraph';
    const STYL = 'readStyle';
	const CTRL = 'readControl';

    /**
     * PhpWord object
     *
     * @var \PhpOffice\PhpWord\PhpWord
     */
    private $phpWord;

    /**
     * Section object
     *
     * @var \PhpOffice\PhpWord\Element\Section
     */
    private $section;

    /**
     * Textrun object
     *
     * @var \PhpOffice\PhpWord\Element\TextRun
     */
    private $textrun;

    /**
     * RTF content
     *
     * @var string
     */
    public $rtf;

    /**
     * Content length
     *
     * @var int
     */
    private $length = 0;

    /**
     * Character index
     *
     * @var int
     */
    private $offset = 0;

    /**
     * Current control word
     *
     * @var string
     */
    private $control = '';

    /**
     * Text content
     *
     * @var string
     */
    private $text = '';

    /**
     * Parsing a control word flag
     *
     * @var bool
     */
    private $isControl = false;

    /**
     * First character flag: watch out for control symbols
     *
     * @var bool
     */
    private $isFirst = false;

	/**
	 * Indicate there is a new paragraph we didn't do addTextRun yet
	 *
	 * When we read a \par, we do addTextRun later, because next output destination might be a TitleRun,
	 * but a TitleRun create a newline too, so if we do addTextRun right before a TitleRun, there would be one
	 * more newline in coverted doc
	 *
	 * @var bool
	 */
	private $pendingParagraph = false;

    /**
     * Group groups
     *
     * @var array
     */
    private $groups = array();

    /**
     * Parser flags; not used
     *
     * @var array
     */
    private $flags = array();

	private $styles = array();
	private $colors = array();
	private $fonts = array();
	private $hexValues = array();

	private $imgCount = 0;
	private $imageBuffer = array();
	private $image = array();
	private $imageWritten = false;

	private $table = null;
	private $row = null;
	private $cell = null;
	private $lastRow = false;
	private $newRow = false;
	private $rowIndex = -1;
	private $inTable = false;

	private $languageMap = array(
		'1028' => 'BIG-5',      //cht
		'1032' => 'ISO-8859-7', //greek
		'1033' => 'CP1252',      //eng
		'1038' => 'UHC',        //hun
		'1041' => 'CP932',      //jpn
		'1049' => 'KOI8-R',     //rus
		'1055' => 'ISO-8859-9', //trk
		'2052' => 'GB18030'     //chs
	);

	private $charsetMap = array(
		'0'   => 'CP1252',     //eng
		'128' => 'CP932',      //jpn
		'129' => 'UHC',        //hun
		'134' => 'GB18030',    //chs
		'136' => 'BIG-5',      //cht
		'161' => 'ISO-8859-7', //greek
		'162' => 'ISO-8859-9', //trk
		'130' => 'EUC-KR',     //krn
		'205' => 'KOI8-R'      //rus
	);

	private $readMethods = array(
		'fldinst' => 'readLinkTarget',
		'fldrslt' => 'readLinkText',
		'bkmkstart' => 'readBookmark',
		'stylesheet' => 'readStyleSheet',
		'fonttbl' => 'readFontTable'
	);

	private $tableCtrlKw = array(
		'trowd',
		'irow',
		'cell',
		'row',
		'cell',
		'lastrow',
		'clvmgf',
		'clvmrg',
		'cellx'
	);

    /**
     * Parse RTF content
     *
     * - Marks controlling characters `{`, `}`, and `\`
     * - Removes line endings
     * - Builds control words and control symbols
     * - Pushes every other character into the text queue
     *
     * @param \PhpOffice\PhpWord\PhpWord $phpWord
     * @return void
     * @todo Use `fread` stream for scalability
     */
    public function read(PhpWord $phpWord)
    {
        $markers = array(
            123 => 'markOpening',   // {
            125 => 'markClosing',   // }
            92  => 'markBackslash', // \
            10  => 'markNewline',   // LF
            13  => 'markNewline'    // CR
        );

        $this->phpWord = $phpWord;
        $this->section = $phpWord->addSection();
        $this->length = strlen($this->rtf);

        $this->flags['paragraph'] = true; // Set paragraph flag from the beginning
		$this->flags['ignoreTotal'] = 1;

		$this->makeNumberStyle();

        // Walk each characters
        while ($this->offset < $this->length) {
            $char  = $this->rtf[$this->offset];
            $ascii = ord($char);

            if (isset($markers[$ascii])) { // Marker found: {, }, \, LF, or CR
                $markerFunction = $markers[$ascii];
                $this->$markerFunction();
            } else {
				if ($this->isControl === false) { // Non control word: Push character
					if ($this->flags['image'] && ('jpeg' === $this->flags['imageType'] || 'png' === $this->flags['imageType'] || 'wmetafile' === $this->flags['imageType'])) {
						$this->pushImage($char);
					} else {
						$this->pushText($char);
					}
                } else {
                    if (preg_match("/^[a-zA-Z0-9-']?$/", $char)) { // No delimiter: Buffer control
                        $this->control .= $char;
                        $this->isFirst = false;
                    } else { // Delimiter found: Parse buffered control
                        if ($this->isFirst) {
                            $this->isFirst = false;
                        } else {
                            if ($char == ' ') { // Discard space as a control word delimiter
                                $this->flushControl(true);
                            }
                        }
                    }
                }
            }
            $this->offset++;
        }
        $this->flushText();
    }

    /**
     * Mark opening braket `{` character.
     *
     * @return void
     */
    private function markOpening()
    {
        $this->flush(true);
		array_push($this->groups, $this->flags);

		if ($this->flags['image']) {
			unset($this->flags['image']);
		}
    }

    /**
     * Mark closing braket `}` character.
     *
     * @return void
     */
    private function markClosing()
    {
        $this->flush(true);
        $this->flags = array_pop($this->groups);
    }

    /**
     * Mark backslash `\` character.
     *
     * @return void
     */
    private function markBackslash()
    {
		$char = $this->rtf[$this->offset + 1];
		$asciiCode = ord($char);
		$nextAsciiCode = ord($this->rtf[$this->offset + 2]);
		// case \ -> escape \
		if (92 === $asciiCode) {
			$this->setControl(false);
			$this->text .= '\\';
			$this->offset++;
		// case ' -> hex value
		} else if (39 === $asciiCode) {
			if ($this->isControl) {
				$this->flushControl(true);
			}
			$this->setControl(false);
			if (true === $this->flags['ignore'] && 0 < $this->flags['ignoreRemain']) {
				$this->flags['ignoreRemain'] -= 1;
				if (0 === $this->flags['ignoreRemain']) {
					$this->flags['ignore'] = false;
				}
			} else {
				array_push($this->hexValues, substr($this->rtf, $this->offset + 2, 2));
			}
			$this->offset += 3;
		// case uXXXX (XXXX is number) -> unicode decimal representation
		// note: not case uc
		} else if (117 === $asciiCode && 48 <= $nextAsciiCode && 57 >= $nextAsciiCode) {
			$this->flush();
			$this->setControl(false);
			$this->offset++;
			$decValue = '';
			$asciiCode = ord($this->rtf[$this->offset + 1]);
			while (48 <= $asciiCode && 57 >= $asciiCode) {
				$decValue .= $this->rtf[$this->offset + 1];
				$this->offset++;
				$asciiCode = ord($this->rtf[$this->offset + 1]);
			}
			$this->flags['ignore'] = true;
			$this->flags['ignoreRemain'] = $this->flags['ignoreTotal'];
			$this->text .= html_entity_decode("&#x" . dechex($decValue) . ";", ENT_COMPAT, 'UTF-8');
		 } else {
			$this->flush();
			$this->setControl(true);
			$this->control = '';
		}
    }

    /**
     * Mark newline character: Flush control word because it's not possible to span multiline.
     *
     * @return void
     */
    private function markNewline()
    {
        if ($this->isControl) {
            $this->flushControl(true);
        }
    }

    /**
     * Flush control word or text.
     *
     * @param bool $isControl
     * @return void
     */
    private function flush($isControl = false)
    {
		if ($this->isControl) {
			$this->flushControl($isControl);
			return;
		}

        if ($this->flags['image'] && $this->imageWritten) {
			$this->flushImage();
			return;
		}

		$this->flushHexValues();
		$this->flushText();
    }

    /**
     * Flush control word.
     *
     * @param bool $isControl
     * @return void
     */
    private function flushControl($isControl = false)
    {
        if (preg_match("/^([A-Za-z]+)(-?[0-9]*) ?$/", $this->control, $match) === 1) {
			if (0 !== strlen($this->text)) {
				$this->flushText();
			}
            list(, $control, $parameter) = $match;
            $this->parseControl($control, $parameter);
        }

        if ($isControl === true) {
            $this->setControl(false);
		}
    }

	private static function _execv($bin, $args)
	{
		$pid = pcntl_fork();
		if ($pid === -1) {
			return FALSE;
		} else if ($pid) {
			pcntl_waitpid($pid, $status);
			if (pcntl_wifexited($status)) {
				return 0 === pcntl_wexitstatus($status);
			}
		} else {
			if (FALSE === pcntl_exec($bin, $args)) {
				exit(1);
			}
		}
		return FALSE;
	}

	private function flushImage()
	{
		if (!$this->imageWritten) {
			return;
		}
		if (!isset($this->fpImg)) {
			return;
		}
		fclose($this->fpImg);
		$this->imageWritten = false;

		if ('wmetafile' === $this->flags['imageType']) {
			if (FALSE === self::_execv('/var/packages/Spreadsheet/target/tools/wmf2svg', array($this->imgPath))) {
				return;
			}
			$imgPath = $this->imgPath . '-1.png';
		} else {
			$imgPath = $this->imgPath;
		}

		unset($this->fpImg);
		unset($this->imgPath);

		$width = \PhpOffice\PhpWord\Shared\Drawing::twipsToPixels(floatval($this->flags['styles']['image']['width']));
		$height = \PhpOffice\PhpWord\Shared\Drawing::twipsToPixels(floatval($this->flags['styles']['image']['height']));

		if (isset($this->flags['styles']['image']['scalew'])) {
			$width = $width * floatval($this->flags['styles']['image']['scalew']) / 100;
		}
		if (isset($this->flags['styles']['image']['scaleh'])) {
			$height = $height * floatval($this->flags['styles']['image']['scaleh']) / 100;
		}

		$style = array(
			'width'        => $width,
			'height'       => $height,
			'synoFileType' => 'image',
			'synoAlt'      => 'img' . $this->imgCount,
			'synoFileId'   => 'img' . $this->imgCount
		);

		$this->section->addImage($imgPath, $style);
	}

    /**
     * Flush text in queue.
     *
     * @return void
     */
    private function flushText()
    {
		if ($this->text != '') {
            if (isset($this->flags['property'])) { // Set property
                $this->flags['value'] = $this->text;
            } else { // Set text
                if ($this->flags['paragraph'] === true) {
                    $this->flags['paragraph'] = false;
                    $this->flags['text'] = $this->text;
                }
            }

			$readMethod = $this->readMethods[$this->flags['property']];
			if (isset($readMethod)) {
				$this->$readMethod();
			} else if (!isset($this->flags['skipped']) || false === $this->flags['skipped']) {
				$this->readText();
			}

            $this->text = '';
        }
    }

    /**
     * Reset control word and first char state.
     *
     * @param bool $value
     * @return void
     */
    private function setControl($value)
    {
        $this->isControl = $value;
        $this->isFirst = $value;
    }

	private function pushImage($char)
	{
		$this->imageBuffer[] = $char;

		if (4 === count($this->imageBuffer)) {
			fwrite($this->fpImg, pack('n', hexdec(join($this->imageBuffer))));
			unset($this->imageBuffer);
			$this->imageBuffer = array();
			$this->imageWritten = true;
		}
	}

    /**
     * Push text into queue.
     *
     * @param string $char
     * @return void
     */
    private function pushText($char)
    {
		if (true === $this->flags['ignore'] && 0 < $this->flags['ignoreRemain']) {
			$this->flags['ignoreRemain'] -= 1;
			if (0 === $this->flags['ignoreRemain']) {
				$this->flags['ignore'] = false;
			}
			return;
		}

		$this->flushHexValues();

        if ($char == '<') {
            $this->text .= "&lt;";
        } elseif ($char == '>') {
            $this->text .= "&gt;";
        } else {
            $this->text .= $char;
        }
	}

	private function flushHexValues()
	{
		$hexValuesCount = count($this->hexValues);
		if (0 === $hexValuesCount) {
			return;
		}

		for ($i = 0; $i + 1 < $hexValuesCount; $i+=2) {
			$packedValue = pack('n', hexdec($this->hexValues[$i] . $this->hexValues[$i+1]));
			$this->text .= mb_convert_encoding($packedValue, "UTF-8", $this->codePage);
		}
		if (1 === $hexValuesCount % 2) {
			$packedValue = pack('n', hexdec($this->hexValues[$hexValuesCount - 1]));
			$this->text .= mb_convert_encoding($packedValue, "UTF-8", "CP1252");
			if ('pntext' === $this->flags['property']) {
				$this->flags['listStyleName'] = 'PHPWordList2';
			}
		}

		unset($this->hexValues);
		$this->hexValues = array();
	}

    /**
     * Parse control.
     *
     * @param string $control
     * @param string $parameter
     * @return void
     */
    private function parseControl($control, $parameter)
    {
        $controls = array(
            'par'       => array(self::PARA,    'paragraph',    true),
            'b'         => array(self::STYL,    'font',         'bold',         $parameter, 'toggle'),
            'i'         => array(self::STYL,    'font',         'italic',       $parameter, 'toggle'),
			'ul'        => array(self::STYL,    'font',         'underline',    $parameter, 'toggle'),
			'ulnone'    => array(self::STYL,    'font',         'underline',    '',         'value'),
			'strike'    => array(self::STYL,    'font',         'strikethrough',$parameter, 'toggle'),
            'fs'        => array(self::STYL,    'font',         'size',         $parameter, 'value', 'divideBy2'),
			'super'     => array(self::STYL,    'font',         'superScript',  true,       'value'),
			'sub'       => array(self::STYL,    'font',         'subScript',    true,       'value'),
			'nosupersub'=> array(self::STYL,    'font',         'nosupersub'),
			'highlight' => array(self::STYL,    'font',         'bgColor',      $parameter, 'color'),
			'cf'        => array(self::STYL,    'font',         'color',        $parameter, 'color'),
			'f'         => array(self::STYL,    'font',         'name',         $parameter),
			'qc'        => array(self::STYL,    'paragraph',    'align',        'center',   'value'),
			'qr'        => array(self::STYL,    'paragraph',    'align',        'right',    'value'),
			'qj'        => array(self::STYL,    'paragraph',    'align',        'both',     'value'),
			'sa'        => array(self::STYL,    'paragraph',    'spaceAfter',   $parameter, 'value'),
			'fi'        => array(self::STYL,    'paragraph',    'hanging',      $parameter, 'value'),
			'li'        => array(self::STYL,    'paragraph',    'indent',       $parameter, 'value'),
			's'         => array(self::STYL,    'paragraph',    'styleIndex',   $parameter, 'value'),
			'picwgoal'  => array(self::STYL,    'image',        'width',        $parameter, 'value'),
			'pichgoal'  => array(self::STYL,    'image',        'height',       $parameter, 'value'),
			'picscalex' => array(self::STYL,    'image',        'scalew',       $parameter, 'value'),
			'picscaley' => array(self::STYL,    'image',        'scaleh',       $parameter, 'value'),
            'fonttbl'   => array(self::CTRL,    'fonttbl',      false),
			'panose'    => array(self::CTRL,    'panose',       true),
            'colortbl'  => array(self::CTRL,    'colortbl',     true),
            'info'      => array(self::CTRL,    'info',         true),
            'generator' => array(self::CTRL,    'generator',    true),
            'title'     => array(self::CTRL,    'title',        true),
            'subject'   => array(self::CTRL,    'subject',      true),
            'category'  => array(self::CTRL,    'category',     true),
            'keywords'  => array(self::CTRL,    'keywords',     true),
            'comment'   => array(self::CTRL,    'comment',      true),
			'shppict'   => array(self::CTRL,    'pic',          true),
            'fldinst'   => array(self::CTRL,    'fldinst',      false),
			'datafield' => array(self::CTRL,    'datafield',    true),
			'fldrslt'   => array(self::CTRL,    'fldrslt',      false),
			'stylesheet'=> array(self::CTRL,    'stylesheet',   false),
			'listtable' => array(self::CTRL,    'listtable',    true),
			'listoverridetable'=> array(self::CTRL, 'listoverridetable', true),
			'fchars'    => array(self::CTRL,    'fchars',       true),
			'lchars'    => array(self::CTRL,    'lchars',       true),
			'rsidtbl'   => array(self::CTRL,    'rsidtbl',      true),
			'wgrffmtfilter'=> array(self::CTRL, 'wgrffmtfilter',true),
			'latentstyles'=> array(self::CTRL, 'latentstyles',  true),
			'bkmkstart' => array(self::CTRL,    'bkmkstart',    false),
			'bkmkend'   => array(self::CTRL,    'bkmkend',      true),
			'pict'      => array(self::CTRL,    'pict',			false),
			'picprop'   => array(self::CTRL,    'picprop',      true),
			'jsksu'     => array(self::CTRL,    'jsksu',        true),
			'jpegblip'  => array(self::CTRL,    'jpeg',         false),
			'wmetafile' => array(self::CTRL,    'wmetafile',    false),
			'blipuid'   => array(self::CTRL,    'blipuid',      true),
			'object'    => array(self::CTRL,    'object',       true),
			'pnseclvl'  => array(self::CTRL,    'pnseclvl',     true),
			'xmlnstbl'  => array(self::CTRL,    'xmlnstbl',     true),
			'xmlopen'   => array(self::CTRL,    'xmlopen',      true),
			'themedata' => array(self::CTRL,    'themedata',    true),
			'colorschememapping'=>array(self::CTRL,'colorschememapping', true),
			'datastore' => array(self::CTRL,    'datastore',    true),
			'ftnsep'    => array(self::CTRL,    'ftnsep',       true),
			'ftnsepc'   => array(self::CTRL,    'ftnsepc',      true),
			'aftnsep'   => array(self::CTRL,    'aftnsep',      true),
			'aftnsepc'  => array(self::CTRL,    'aftnsepc',     true),
			'pntxtb'    => array(self::CTRL,    'pntxtb',       true),
			'falt'      => array(self::CTRL,    'falt',         true),
			'header'    => array(self::CTRL,    'header',       true),
			'headerl'   => array(self::CTRL,    'headerl',      true),
			'headerr'   => array(self::CTRL,    'headerr',      true),
			'headerf'   => array(self::CTRL,    'headerf',      true),
			'footer'    => array(self::CTRL,    'footer',       true),
			'footerl'   => array(self::CTRL,    'footerl',      true),
			'footerr'   => array(self::CTRL,    'footerr',      true),
			'footerf'   => array(self::CTRL,    'footerf',      true),
			'pntext'    => array(self::CTRL,    'pntext',       false)
		);

		if ('deflang' === $control) {
			$this->defCodePage = $this->languageMap[$parameter];
			return;
		}

		if ('langfe' === $control || 'lang' === $control) {
			if (isset($this->languageMap[$parameter])) {
				$this->codePage = $this->languageMap[$parameter];
			} else {
				$this->codePage = $this->defCodePage;
			}
			return;
		}

		if ('uc' === $control) {
			$this->flags['ignoreTotal'] = (int)$parameter;
			return;
		}

		if ('pard' === $control) {
			$this->flags['styles']['paragraph'] = array();
			$this->flags['styles']['font']['bgColor'] = '';
			return;
		}

		if ('plain' === $control) {
			$this->codePage = $this->defCodePage;
			$this->flags['styles']['font'] = array();
			return;
		}

		if ('colortbl' === $this->flags['property']) {
			$this->flags[$control] = (10 > intval($parameter)) ? '0' . dechex($parameter) : dechex($parameter);
			if ('blue' === $control) {
				$this->colors[] = $this->flags['red'] . $this->flags['green'] . $this->flags['blue'];
			}
			return;
		}

		if ('fonttbl' === $this->flags['property'] && 'f' === $control) {
			$this->flags['fontIndex'] = $parameter;
			return;
		}

		if ('fcharset' === $control) {
			$index = $this->flags['fontIndex'];
			$this->codePage = $this->charsetMap[$parameter];
			$this->fonts[$index]['codePage'] = $this->charsetMap[$parameter];
			return;
		}

        if (isset($controls[$control])) {
            list($function) = $controls[$control];
            if (method_exists($this, $function)) {
                $directives = $controls[$control];
                array_shift($directives); // remove the function variable; we won't need it
                $this->$function($directives);
			}
		}
    }

    /**
     * Read paragraph.
     *
     * @param array $directives
     * @return void
     */
    private function readParagraph($directives)
    {
		if (true === $this->flags['skipped']) {
			return;
		}
        list($property, $value) = $directives;
		if (true === $this->pendingParagraph) {
			$this->textrun = $this->section->addTextRun();
		}
		$this->pendingParagraph = true;
        $this->flags[$property] = $value;

		$this->titlerun = null;
		$this->bookmark = '';
    }

    /**
     * Read style.
     *
     * @param array $directives
     * @return void
     */
    private function readStyle($directives)
    {
		list($style, $property, $value, $type, $converter) = $directives;
		if ('toggle' === $type) {
			if ('underline' === $property) {
				$value = (!isset($value) || '0' !== $value) ? 'single' : '';
			} else {
				$value = (!isset($value) || '0' !== $value) ? true : false;
			}
		}
		if (isset($converter)) {
			$value = $this->$converter($value);
		}
		if ('indent' === $property || 'hanging' === $property) {
			$value = abs(intval($value));
		}
		if ('nosupersub' === $property) {
			$this->flags['styles'][$style]['superScript'] = false;
			$this->flags['styles'][$style]['subScript'] = false;
		}
		if ('color' === $type) {
			if ('0' === $value) {
				$value = '';
			} else {
				$value = $this->colors[intval($value) - 1];
			}
		}
		if ('name' === $property) {
			$index = $value;
			$value = $this->fonts[$index]['name'];
			$this->codePage = $this->fonts[$index]['codePage'];
		}
		$this->flags['styles'][$style][$property] = $value;
    }

    /**
     * Read skip.
     *
     * @param array $directives
     * @return void
     */
    private function readControl($directives)
    {
        list($property, $skip) = $directives;

		$this->flags['property'] = $property;
		$this->flags['skipped'] = $skip;

		global $synoTempPath;
		switch ($property) {
			case 'pict':
				$this->flags['image'] = true;
				$this->imgCount++;

				$this->imgPath = dirname(dirname($synoTempPath)) . '/' . $this->fileName . '-img' . $this->imgCount;
				$this->fpImg = fopen($this->imgPath, 'w');
				break;
			case 'jpeg':
			case 'png':
			case 'wmetafile':
				$this->flags['imageType'] = $property;
				break;
			default:
				break;
		}
    }

    /**
     * Read text.
     *
     * @return void
     */
    private function readText()
    {
		$paragraphStyle = $this->flags['styles']['paragraph'];
		if (isset($paragraphStyle['styleIndex']) && 
			isset($this->styles[$paragraphStyle['styleIndex']]) &&
			isset($this->styles[$paragraphStyle['styleIndex']]['level'])) {
			if (!isset($this->titlerun)) {
				$this->titlerun = $this->section->addTitleRun($paragraphStyle, $this->styles[$paragraphStyle['styleIndex']]['level'], $this->bookmark, $this->phpWord);
				$this->bookmark = '';
				$this->pendingParagraph = false;
			}
			$this->titlerun->addText($this->text, $this->flags['styles']['font'], $paragraphStyle);
		} else {
			if ('pntext' === $this->flags['property']) {
				$listStyleName = isset($this->flags['listStyleName']) ? $this->flags['listStyleName'] : 'PHPWordList1';
				$this->listItemRun = $this->section->addListItemRun(0, $listStyleName, array());
				$this->textrun = $this->listItemRun->addTextRun();
				$this->pendingParagraph = false;
			} else if (!isset($this->textrun) || true === $this->pendingParagraph) {
				$this->textrun = $this->section->addTextRun();
				$this->pendingParagraph = false;
			}
			$this->textrun->setParagraphStyle($this->flags['styles']['paragraph']);
			if ('pntext' !== $this->flags['property']) {
				$text = $this->textrun->addText($this->text);
				if (isset($this->flags['styles']['font'])) {
					$text->getFontStyle()->setStyleByArray($this->flags['styles']['font']);
				}
			}
		}
    }

	private function readStyleSheet()
	{
		$styles = $this->flags['styles'];
		$index = $styles['paragraph']['styleIndex'];
		if (!isset($index)) {
			return;
		}

		$this->styles[$index] = $styles;
		$name = trim($this->text, ';');

		$this->styles[$index]['name'] = $name;
		if (preg_match('/heading (\d+)/', $name, $matches)) {
			$this->styles[$index]['level'] = $matches[1];
		}
	}

	private function readFontTable()
	{
		$index = $this->flags['fontIndex'];
		if (!isset($index)) {
			return;
		}
		$name = trim($this->text, ';');
		$name = trim($name, '@');

		if (0 === strlen($name)) {
			return;
		}
		$this->fonts[$index]['name'] = $name;
	}

	private function readLinkTarget()
	{
		if (preg_match('/^HYPERLINK \\\\l \"(.*)\"$/', $this->text, $matches)) {
			$target = '#' . $matches[1];
		} else if (preg_match('/^HYPERLINK \"(.*)\"$/', $this->text, $matches)) {
			$target = $matches[1];
		} else {
			return;
		}
		if (!isset($this->textrun) || true === $this->pendingParagraph) {
			$this->textrun = $this->section->addTextRun();
			$this->pendingParagraph = false;
		}
		$this->textrun->setParagraphStyle($this->flags['styles']['paragraph']);
		$this->link = $this->textrun->addLink($target, '', $fontStyle, $paragraphStyle, false);
	}

	private function readLinkText()
	{
		if (!isset($this->link)) {
			return;
		}
		$this->link->addText($this->text, $this->flags['styles']['font'], $this->flags['styles']['paragraph']);
	}

	private function readBookmark()
	{
		$this->bookmark = $this->text;
	}

	private function readTable($control, $parameter)
	{
		if ('lastrow' == $control) {
			if (true == $this->lastrow) {
				$this->inTable = false;
				$this->table = null;
				$this->rowIndex = -1;
			} else {
				$this->lastrow = true;
			}
			return;
		}

		if (!isset($this->table)) {
			$this->table = array('cellStyle' => array(), 'rows' => array());
			$this->inTable = true;
		}

		if ('trowd' == $control) {
			if (false == $this->newRow) {
				$this->newRow = true;
				$this->rowIndex++;
				$this->table->rows[] = array('cells' => array());
			} else {
				$this->newRow = false;
			}
			return;
		}

		if ($this->newRow && 'cellx' == $control) {
			if(!in_array($parameter, $this->table->cellStyle)) {
				$this->table->cellStyle[] = $parameter;
			}
		}

		if ('cell' == $control) {
			$this->textrun = null;
		}
	}

	private function divideBy2($value)
	{
		return $value / 2;
	}

	private function makeNumberStyle()
	{
		$style = array('levels' => array(), 'type' => 'hybridMultilevel', 'numId' => 1);
		for ($i = 0; $i < 10; $i++) {
			$level = array();
			$level['level'] = $i;
			$level['start'] = 1;
			$level['format'] = 'decimal';
			$level['text'] = "%$i.";
			$level['align'] = 'left';

			$style['levels'][$i] = $level;
		}
		$this->phpWord->addNumberingStyle("PHPWordList1", $style);

		$style = array('levels' => array(), 'type' => 'hybridMultilevel', 'numId' => 2);
		for ($i = 0; $i < 10; $i++) {
			$level = array();
			$level['level'] = $i;
			$level['start'] = 1;
			$level['format'] = 'bullet';
			$level['text'] = '·';
			$level['align'] = 'left';

			$style['levels'][$i] = $level;
		}
		$this->phpWord->addNumberingStyle("PHPWordList2", $style);
	}
}
