<?php
/**
 * This file is part of PHPWord - A pure PHP library for reading and writing
 * word processing documents.
 *
 * PHPWord is free software distributed under the terms of the GNU Lesser
 * General Public License version 3 as published by the Free Software Foundation.
 *
 * For the full copyright and license information, please read the LICENSE
 * file that was distributed with this source code. For the full list of
 * contributors, visit https://github.com/PHPOffice/PHPWord/contributors.
 *
 * @link        https://github.com/PHPOffice/PHPWord
 * @copyright   2010-2014 PHPWord contributors
 * @license     http://www.gnu.org/licenses/lgpl.txt LGPL version 3
 */

namespace PhpOffice\PhpWord\Reader\Word2007;

use PhpOffice\PhpWord\PhpWord;
use PhpOffice\PhpWord\Shared\XMLReader;
use PhpOffice\PhpWord\Shared\Drawing;
use PhpOffice\PhpWord\Shared\Converter;
use PhpOffice\PhpWord\Style;
use PhpOffice\PhpWord\Style\Font as FontStyle;

/**
 * Abstract part reader
 *
 * This class is inherited by ODText reader
 *
 * @since 0.10.0
 */
abstract class AbstractPart
{
    /**
     * Conversion method
     *
     * @const int
     */
    const READ_VALUE = 'attributeValue';            // Read attribute value
    const READ_EQUAL = 'attributeEquals';           // Read `true` when attribute value equals specified value
    const READ_TRUE  = 'attributeTrue';             // Read `true` when element exists
    const READ_FALSE = 'attributeFalse';            // Read `false` when element exists
    const READ_SIZE  = 'attributeMultiplyByTwo';    // Read special attribute value for Font::$size

    /**
     * Document file
     *
     * @var string
     */
    protected $docFile;

    /**
     * XML file
     *
     * @var string
     */
    protected $xmlFile;

    /**
     * Part relationships
     *
     * @var array
     */
    protected $rels = array();

    /**
     * root object
     */
    protected $rootObject;

    /**
     * Read part.
     */
    abstract public function read(PhpWord $phpWord);

    /**
     * Create new instance
     *
     * @param string $docFile
     * @param string $xmlFile
     */
    public function __construct($docFile, $xmlFile)
    {
        $this->docFile = $docFile;
        $this->xmlFile = $xmlFile;
    }

    /**
     * Set relationships.
     *
     * @param array $value
     * @return void
     */
    public function setRels($value)
    {
        $this->rels = $value;
    }

    private function checkHeadingLevel($paragraphStyle)
    {
        if ( !empty($paragraphStyle['outlineLevel']) ) {
            return intval($paragraphStyle['outlineLevel']) + 1;
        }
        if ( !empty($paragraphStyle['styleName']) ) {
            $gStyle = Style::getStyle($paragraphStyle['styleName']);
            if ( method_exists($gStyle, 'getParagraphStyle') ) {
                $gStyle = $gStyle->getParagraphStyle();
            }
            if ( !empty($gStyle) ) {
                if ( !empty($finalStyle = $gStyle->getFinalStyle()) ) {
                    $outlineLevel = $finalStyle->getOutlineLevel();
                    if ( !is_null($outlineLevel) ) { return intval($outlineLevel) + 1; }
                }
            }
        }
        return -1;
    }

    private function checkNumbering(&$paragraphStyle)
    {
        $style = Array();
        $numStyle = $paragraphStyle['numStyle'];
        $numLevel = $paragraphStyle['numLevel'];
        unset($paragraphStyle['numStyle']);
        unset($paragraphStyle['numLevel']);

        $style['numStyle'] = $numStyle;
        if ( !is_null($numStyle) && is_null(Style::getStyle("PHPWordListMSWord{$numStyle}")) )
            return null;
        $style['numLevel'] = is_null($numLevel) ? 0 : $numLevel;
        if ( is_null($numStyle) && !empty($paragraphStyle['styleName']) ) {
            $gStyle = Style::getStyle($paragraphStyle['styleName']);
            if ( method_exists($gStyle, 'getParagraphStyle') ) {
                $gStyle = $gStyle->getParagraphStyle();
            }
            if ( !empty($gStyle) ) {
                if ( !empty($finalStyle = $gStyle->getFinalStyle()) ) {
                    $style['numStyle'] = $finalStyle->getNumStyle();
                    $style['numLevel'] = $finalStyle->getNumLevel();
                    if ( is_null($numLevel) ) $style['numLevel'] = 0;
                }
            }
        }
        return $style;
    }

    private function appendGlobalStyle(&$styles)
    {
        if ($this->phpWord !== null) {
            $gStyles = Style::getStyle($styles['styleName']);
            if (!is_null($gStyles)) {
                if (method_exists($gStyles, 'getSize')) {
                    $styles['size'] = $gStyles->getSize();
                }
                if (method_exists($gStyles, 'getColor')) {
                    $styles['color'] = $gStyles->getColor();
                }
                if (method_exists($gStyles, 'getIndent') && !isset($styles['indent'])) {
                    $styles['indent'] = $gStyles->getIndent() / 720;
                }
                if (method_exists($gStyles, 'getHanging') && !isset($styles['hanging'])) {
                    $styles['hanging'] = $gStyles->getHanging() / 720;
                }
                if (method_exists($gStyles, 'getFirstLine') && !isset($styles['firstLine'])) {
                    $styles['firstLine'] = $gStyles->getFirstLine() / 720;
                }
            }
        }
    }

    protected function storeRevision($xmlReader, $node)
    {
        global $comment;

        $step = function ($node, $op) {
            if ($node->parentNode->nodeName !== 'w:trPr') {
                return 'replace'; // TODO
            }
            switch ($op) {
                case 'del': return 'removeTableRow';
                case 'ins': return 'addTableRow';
                default: return 'replace'; // TODO
            }
        };

        $id = $xmlReader->getAttribute('w:id', $node);
        $op = substr($node->nodeName, 2);
        if (!is_array($comment)) {
            $comment = array();
        }
        if (array_key_exists($id, $comment)) {
            // TODO
        } else {
            $author = $xmlReader->getAttribute('w:author', $node);
            $date = $xmlReader->getAttribute('w:date', $node);
            $text = json_encode(array(
                'op' => array('del' => 'deletion', 'ins' => 'insertion')[$op],
                'step' => $step($node, $op),
            ));
            $comment[$id] = array(
                'type' => 1,
                'id' => $id,
                'author' => $author,
                'date' => $date,
                'text' => $text,
            );
        }

        return array('id' => $id, 'op' => $op);
    }

    /**
     * Read w:p.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @param mixed $parent
     * @param string $docPart
     * @return void
     *
     * @todo Get font style for preserve text
     */
    protected function readParagraph(XMLReader $xmlReader, \DOMElement $domNode, $parent, $docPart = 'document')
    {
        $phpword = $parent->getPhpWord();
        $runNode = null;

        // Paragraph style
        $paragraphStyle = null;
        $numbering = null;
        $heading_lvl = -1;

        if ( $xmlReader->elementExists('w:pPr', $domNode) ) {
            $paragraphStyle = $this->readParagraphStyle($xmlReader, $domNode);
            $this->appendGlobalStyle($paragraphStyle);
            $heading_lvl = $this->checkHeadingLevel($paragraphStyle);
            $numbering = $this->checkNumbering($paragraphStyle);
        }

        if ( $xmlReader->elementExists('w:pPr/w:rPr', $domNode) ) {
            $pNode = $xmlReader->getElement('w:pPr', $domNode);
            $fontStyle = $this->readFontStyle($xmlReader, $pNode);
            if ( isset($fontStyle['size']) ) {
                $paragraphStyle['size'] = $fontStyle['size'];
            }
            $paragraphStyle['synoListItem'] = $fontStyle;
        }

        // Read revision
        $revNode = $xmlReader->getElement('w:pPr/w:rPr/w:del|w:pPr/w:rPr/w:ins', $domNode);
        if ( $revNode ) {
            $paragraphStyle['revision'] = $this->storeRevision($xmlReader, $revNode);
        }

        // ListItem
        // Remark: always put ListItemRun as the outermost container if it exists
        if ( !empty($numbering['numStyle']) ) {
            // List item
            $numId = $numbering['numStyle'];
            $levelId = $numbering['numLevel'];

            // Get indent from numbering style
            $nStyle = Style::getStyle("PHPWordListMSWord{$numId}");
            if ( method_exists($nStyle, 'getLevels') ) {
                $level = $nStyle->getLevels()[$levelId];
                if ( $level ) {
                    $left = $level->getLeft();
                    if ($left) {
                        $paragraphStyle['indent'] = $left;
                    }
                    $hanging = $level->getHanging();
                    if ($hanging) {
                        $paragraphStyle['hanging'] = $hanging;
                    }
                    if ( $level->getFormat() !== 'none' ) {
                        $runNode = $parent->addListItemRun($levelId, "PHPWordListMSWord{$numId}", $paragraphStyle);

                        // Google Doc adds padding-left to list items to exported .docx files
                        // but indent should not be applied to list items
                        $paragraphStyle['indent'] = null;
                    }
                }
            }
        }

        // Heading
        if ( $heading_lvl >= 0 ) {
            $synohead = $xmlReader->getAttribute('w:name', $domNode, 'w:bookmarkStart');
            if ( is_null($runNode) ) { $runNode = $parent; }
            $runNode = $runNode->addTitleRun($paragraphStyle, $heading_lvl, $synohead, $phpword);
        }

        // Text and TextRun
        if ( is_null($runNode) ) { $runNode = $parent; }
        if ( $xmlReader->countElements('w:r|w:hyperlink|w:del|w:ins', $domNode) === 0 ) {
            $runNode->addTextBreak(1, null, $paragraphStyle);
        } else {
            if ( ($runNode instanceof \PhpOffice\PhpWord\Element\TOCRun)
                    || !($runNode instanceof \PhpOffice\PhpWord\Element\TextRun) ) {
                $runNode = $runNode->addTextRun($paragraphStyle);
            }
            $this->readRunsRecursively($xmlReader, $domNode, $runNode, $docPart, $paragraphStyle);
        }
    }

    protected function readRunsRecursively(XMLReader $xmlReader, $domNode, $parent, $docPart, $paragraphStyle = null, $revision = null)
    {
        global $synocomment;
        $synodeco = &$this->phpWord->synoDecorator;
        $inFieldChar = false;
        $ignoreText = false;
        $textContent = '';
        $nodes = $xmlReader->getElements('w:r|w:hyperlink|w:commentRangeStart|w:commentRangeEnd|w:del|w:ins', $domNode);
        foreach ($nodes as $node) {
        // field code
            // FIXME: nested field codes are not supported
            if ( $xmlReader->elementExists('w:fldChar|w:instrText', $node) ) {
                foreach ($xmlReader->getElements('*', $node) as $child) {
                    if ( $child->nodeName == 'w:fldChar' ) {
                        if ( $xmlReader->getAttribute('w:fldCharType', $child) == 'begin' ) {
                            $ignoreText = true;
                        } else {
                            if ( $xmlReader->getAttribute('w:fldCharType', $child) == 'end' ) {
                                $inFieldChar = false;
                            }
                            $ignoreText = false;
                        }
                    } else {
                        // detect page number
                        if (str_replace(' ', '', $child->textContent) == 'PAGE\*MERGEFORMAT') {
                            $inFieldChar = true;
                            if ( false == $synodeco['num-show'] ) {
                                $synodeco['num-show']  = true;
                                $synodeco['num-cover']  = true;
                            } else if ( $docPart != 'document' && $this->rootObject->getType() == 'first' ) {
                                $synodeco['num-cover']  = false;
                            }

                            if ( $this->rootObject instanceof \PhpOffice\PhpWord\Element\Header ) {
                                $synodeco['num-pos']  = 'tr';
                            } else if ( $this->rootObject instanceof \PhpOffice\PhpWord\Element\Footer ) {
                                $synodeco['num-pos']  = 'br';
                            }
                        }
                        // PreserveText
                        $instrText = $xmlReader->getValue('w:instrText', $child);
                        if ( !is_null($instrText) ) {
                            $textContent .= '{' . $instrText . '}';
                        } else {
                            // FIXME: w:fldChar/w:fldData should not be ignored
                            if ( $ignoreText === false ) {
                                $textContent .= $xmlReader->getValue('w:t', $child);
                            }
                        }
                    }
                }
            } else if ( !$inFieldChar || $docPart == 'document' ) {
                switch ( $node->nodeName ) {
                    case 'w:commentRangeStart': {
                        $id = $xmlReader->getAttribute('w:id', $node);
                        $index = array_search($id, $synocomment);
                        if ($index === false) {
                            $synocomment[] = $id;
                        }
                        break;
                    }
                    case 'w:commentRangeEnd': {
                        $id = $xmlReader->getAttribute('w:id', $node);
                        $index = array_search($id, $synocomment);
                        if ($index !== false) {
                            array_splice($synocomment, $index, 1);
                        }
                        break;
                    }
                    case 'w:del':
                        // fall through
                    case 'w:ins': {
                        $this->readRunsRecursively($xmlReader, $node, $parent, $docPart, $paragraphStyle, $this->storeRevision($xmlReader, $node));
                        break;
                    }
                    default: {
                        $this->readRun($xmlReader, $node, $parent, $docPart, $paragraphStyle, $revision);
                        break;
                    }
                }
            }
            try {
                if ( !empty($textContent) ) {
                    $parent->addPreserveText($textContent, $fontStyle, $paragraphStyle);
                }
            } catch (\Exception $e) {
                echo "Warning: Failed to add preserve text: $textContent\n";
            }
        }
    }

    /**
     * Read w:r.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @param mixed $parent
     * @param string $docPart
     * @param mixed $paragraphStyle
     * @return void
     *
     * @todo Footnote paragraph style
     */
    protected function readRun(XMLReader $xmlReader, \DOMElement $domNode, $parent, $docPart, $paragraphStyle = null, $revision = null)
    {
        global $synocomment;

        if (!in_array($domNode->nodeName, array('w:r', 'w:hyperlink'))) {
            return;
        }
        $fontStyle = $this->readFontStyle($xmlReader, $domNode);
        if ( isset($paragraphStyle['styleName']) && !isset($fontStyle['styleName']) ) {
            $fontStyle['styleName'] = $paragraphStyle['styleName'];
        }

        $fontStyle['synocomment'] = $synocomment;
        $fontStyle['revision'] = $revision;
        if ( !is_null($paragraphStyle['fontStyle']) ) {
            unset($paragraphStyle['fontStyle']['styleName']); // works as Google Docs and M$ Word
            foreach( $paragraphStyle['fontStyle'] as $key => $value ) {
                if ( !isset($fontStyle[$key]) ) {
                    $fontStyle[$key] = $value;
                }
            }
            $rFont = new \PhpOffice\PhpWord\Style\Font();
            $rFont->setStyleByArray($fontStyle);
            $rFont = $rFont->getFinalStyle();

            $fontStyle = $rFont->getFinalStyle();
        }

        // skip if it is a hidden element
        if ($xmlReader->getAttribute('syno-hidden', $domNode)) {
            return;
        }

        // Link
        if ($domNode->nodeName == 'w:hyperlink') {
            $rId = $xmlReader->getAttribute('r:id', $domNode);
            $internal = false;
            $target = $this->getMediaTarget($docPart, $rId);
            if (empty($target) && $xmlReader->getAttribute('w:anchor', $domNode)) {
                $internal = true;
                $target = $xmlReader->getAttribute('w:anchor', $domNode);
            }
            if (is_null($target)) {
                return;
            }
            /** Does not matter?
            $textContent = $xmlReader->getValue('w:r/w:t', $domNode);
            */
            $textContent = '';
            $link = $parent->addLink($target, $textContent, $fontStyle, $paragraphStyle, $internal);
            $nodes = $xmlReader->getElements('w:r', $domNode);
            foreach ($nodes as $node) {
                $this->readRun(
                    $xmlReader,
                    $node,
                    $link,
                    $docPart,
                    $paragraphStyle
                );
            }
        } else {
            // Footnote
            if ($xmlReader->elementExists('w:footnoteReference', $domNode)) {
                $parent->addFootnote();

            // Endnote
            } elseif ($xmlReader->elementExists('w:endnoteReference', $domNode)) {
                $parent->addEndnote();

            // Image
            } elseif ($xmlReader->elementExists('w:pict', $domNode)) {
                $rId = $xmlReader->getAttribute('r:id', $domNode, 'w:pict/v:shape/v:imagedata');
                $styles = $this->readImageStyle($xmlReader, $domNode);
                if (!array_key_exists('synoFileType', $styles)) {
                    $styles['synoFileType'] = 'image';
                }
                if (!array_key_exists('synoFileId', $styles)) {
                    $styles['synoFileId'] = $rId;
                }
                if (!array_key_exists('synoAlt', $styles)) {
                    $styles['synoAlt'] = $xmlReader->getAttribute('o:title', $domNode, 'w:pict/v:shape/v:imagedata');
                }
                $target = $this->getMediaTarget($docPart, $rId);
                if (!is_null($target)) {
                    $imageSource = "zip://{$this->docFile}#{$target}";
                    try {
                        $image = $parent->addImage($imageSource, $styles);
                        $image->setAnnotations($fontStyle);
                    } catch (\Exception $e) {
                    }
                }

            // Object
            } elseif ($xmlReader->elementExists('w:object', $domNode)) {
                $rId = $xmlReader->getAttribute('r:id', $domNode, 'w:object/o:OLEObject');
                // $rIdIcon = $xmlReader->getAttribute('r:id', $domNode, 'w:object/v:shape/v:imagedata');
                $target = $this->getMediaTarget($docPart, $rId);
                if (!is_null($target)) {
                    $objSource = "zip://{$this->docFile}#{$target}";
                    try {
                        $parent->addObject($objSource);
                    } catch (\Exception $e) {
                        // echo "Warning: Failed to add object $objSource\n";
                    }
                }

            // Drawing
            } elseif ($xmlReader->elementExists('w:drawing', $domNode)) {
                $this->readDrawing($xmlReader, $xmlReader->getElement('w:drawing', $domNode), $parent, $docPart, $paragraphStyle, $fontStyle);

            } elseif ($xmlReader->elementExists('mc:AlternateContent/mc:Choice/w:drawing', $domNode)) {
                $this->readDrawing($xmlReader, $xmlReader->getElement('mc:AlternateContent/mc:Choice/w:drawing', $domNode), $parent, $docPart, $paragraphStyle, $fontStyle);

            // TextRun
            } else {
                foreach ($xmlReader->getElements('*', $domNode) as $node) {
                    if ($node->nodeName === 'w:br') {
                        if ($xmlReader->getAttribute('w:type', $node) !== 'page') {
                            $parent->addTextBreak(1, null, $paragraphStyle);
                        }
                    } else if ($node->nodeName === 'w:tab') {
                        $parent->addText("    ", $fontStyle, $paragraphStyle);
                    } else if (($node->nodeName === 'w:t') || ($node->nodeName === 'w:delText')) {
                        $textContent = htmlspecialchars($node->nodeValue);
                        $parent->addText($textContent, $fontStyle, $paragraphStyle);
                    }
                }
            }
        }
    }

    /**
     * Read w:tbl.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @param mixed $parent
     * @param string $docPart
     * @return void
     */
    protected function readTable(XMLReader $xmlReader, \DOMElement $domNode, $parent, $docPart = 'document')
    {
        // Table style
        $tblStyle = null;
        if ($xmlReader->elementExists('w:tblPr', $domNode)) {
            $tblStyle = $this->readTableStyle($xmlReader, $domNode);
        }

        /** @var \PhpOffice\PhpWord\Element\Table $table Type hint */
        $table = $parent->addTable($tblStyle);
        foreach ($domNode->childNodes as $tblNode) {
            if ($tblNode->nodeName == 'w:tblGrid') { // Column
                $colgroup = array();
                foreach ($xmlReader->getElements('w:gridCol', $tblNode) as $gridCol) {
                    $colgroup[] = array(
                        'width' => (int)$xmlReader->getAttribute('w:w', $gridCol));
                }
                $table->setColgroup($colgroup);
            } elseif ($tblNode->nodeName == 'w:tr') { // Row
                $rowHeight = $xmlReader->getAttribute('w:val', $tblNode, 'w:trPr/w:trHeight');
                $rowHRule = $xmlReader->getAttribute('w:hRule', $tblNode, 'w:trPr/w:trHeight');
                $rowHRule = $rowHRule == 'exact' ? true : false;
                $rowStyle = array(
                    'tblHeader' => $xmlReader->elementExists('w:trPr/w:tblHeader', $tblNode),
                    'cantSplit' => $xmlReader->elementExists('w:trPr/w:cantSplit', $tblNode),
                    'exactHeight' => $rowHRule,
                );

                $row = $table->addRow($rowHeight, $rowStyle);
                foreach ($tblNode->childNodes as $rowNode) {
                    if ($rowNode->nodeName == 'w:trPr') { // Row style
                        // @todo Do something with row style
                        $revNode = $xmlReader->getElement('w:del|w:ins', $rowNode);
                        if ($revNode) {
                            $row->setRevision($this->storeRevision($xmlReader, $revNode));
                        }
                    } elseif ($rowNode->nodeName == 'w:tc') { // Cell
                        $cellWidth = $xmlReader->getAttribute('w:w', $rowNode, 'w:tcPr/w:tcW');
                        $cellStyle = null;
                        $cellStyleNode = $xmlReader->getElement('w:tcPr', $rowNode);
                        if (!is_null($cellStyleNode)) {
                            $cellStyle = $this->readCellStyle($xmlReader, $cellStyleNode);
                        }

                        $cell = $row->addCell($cellWidth, $cellStyle);
                        foreach ($rowNode->childNodes as $cellNode) {
                            if ($cellNode->nodeName == 'w:p') { // Paragraph
                                $this->readParagraph($xmlReader, $cellNode, $cell, $docPart);
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Read w:pPr.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @return array|null
     */
    protected function readParagraphStyle(XMLReader $xmlReader, \DOMElement $domNode)
    {
        if (!$xmlReader->elementExists('w:pPr', $domNode)) {
            return null;
        }

        $styleNode = $xmlReader->getElement('w:pPr', $domNode);
        $styleDefs = array(
            'styleName'       => array(self::READ_VALUE, 'w:pStyle'),
            'align'           => array(self::READ_VALUE, 'w:jc'),
            'next'            => array(self::READ_VALUE, 'w:next'),
            'outlineLevel'    => array(self::READ_VALUE, 'w:outlineLvl', 'w:val'),
            'indent'          => array(self::READ_VALUE, 'w:ind', 'w:left'),
            'hanging'         => array(self::READ_VALUE, 'w:ind', 'w:hanging'),
            'firstLine'       => array(self::READ_VALUE, 'w:ind', 'w:firstLine'),
            'spaceAfter'      => array(self::READ_VALUE, 'w:spacing', 'w:after'),
            'spaceBefore'     => array(self::READ_VALUE, 'w:spacing', 'w:before'),
            'widowControl'    => array(self::READ_FALSE, 'w:widowControl'),
            'keepNext'        => array(self::READ_TRUE,  'w:keepNext'),
            'keepLines'       => array(self::READ_TRUE,  'w:keepLines'),
            'pageBreakBefore' => array(self::READ_TRUE,  'w:pageBreakBefore'),
            'numStyle'        => array(self::READ_VALUE, 'w:numPr/w:numId', 'w:val'),
            'numLevel'        => array(self::READ_VALUE, 'w:numPr/w:ilvl', 'w:val'),
            'borderLeftSize'  => array(self::READ_VALUE, 'w:pBdr/w:left', 'w:sz'),
            'borderLeftStyle' => array(self::READ_VALUE, 'w:pBdr/w:left', 'w:val'),
            'borderRightSize' => array(self::READ_VALUE, 'w:pBdr/w:right', 'w:sz'),
            'borderRightStyle'=> array(self::READ_VALUE, 'w:pBdr/w:right', 'w:val'),
            'borderTopSize'   => array(self::READ_VALUE, 'w:pBdr/w:top', 'w:sz'),
            'borderTopStyle'  => array(self::READ_VALUE, 'w:pBdr/w:top', 'w:val'),
            'borderBottomSize'=> array(self::READ_VALUE, 'w:pBdr/w:bottom', 'w:sz'),
            'borderBottomStyle'=> array(self::READ_VALUE, 'w:pBdr/w:bottom', 'w:val'),
        );

        return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs);
    }

    /**
     * Read w:rPr
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @return array|null
     */
    protected function readFontStyle(XMLReader $xmlReader, \DOMElement $domNode)
    {
        if (is_null($domNode)) {
            return null;
        }
        // Hyperlink has an extra w:r child
        if ($domNode->nodeName == 'w:hyperlink') {
            $domNode = $xmlReader->getElement('w:r', $domNode);
        }
        if (!$xmlReader->elementExists('w:rPr', $domNode)) {
            return null;
        }

        $styleNode = $xmlReader->getElement('w:rPr', $domNode);
        $styleDefs = array(
            'styleName'           => array(self::READ_VALUE, 'w:rStyle'),
            'name'                => array(self::READ_VALUE, 'w:rFonts', 'w:ascii'),
            'hint'                => array(self::READ_VALUE, 'w:rFonts', 'w:hint'),
            'size'                => array(self::READ_SIZE,  'w:sz'),
            'color'               => array(self::READ_VALUE, 'w:color'),
            'underline'           => array(self::READ_VALUE, 'w:u'),
            'bold'                => array(self::READ_TRUE,  'w:b'),
            'italic'              => array(self::READ_TRUE,  'w:i'),
            'strikethrough'       => array(self::READ_TRUE,  'w:strike'),
            'doubleStrikethrough' => array(self::READ_TRUE,  'w:dstrike'),
            'smallCaps'           => array(self::READ_TRUE,  'w:smallCaps'),
            'allCaps'             => array(self::READ_TRUE,  'w:caps'),
            'superScript'         => array(self::READ_EQUAL, 'w:vertAlign', 'w:val', 'superscript'),
            'subScript'           => array(self::READ_EQUAL, 'w:vertAlign', 'w:val', 'subscript'),
            'fgColor'             => array(self::READ_VALUE, 'w:highlight'),
            'bgColor'             => array(self::READ_VALUE, 'w:shd', 'w:fill'),
            'rtl'                 => array(self::READ_TRUE,  'w:rtl'),
        );

        return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs);
    }


    /**
     * Read image style
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @return array
     */
    protected function readImageStyle(XMLReader $xmlReader, \DOMElement $domNode)
    {
        $arr = array();
        $style = $xmlReader->getAttribute('style', $domNode, 'w:pict/v:shape');
        $styleattr = explode(';', $style);
        foreach ($styleattr as $attr) {
            if (strpos($attr, ':')) {
                list($k, $v) = explode(':', $attr);
                $k = trim($k);
                switch ($k) {
                    case 'width':
                        $arr[$k] = $this->convertToPx($v);
                        break;
                    case 'height':
                        $arr[$k] = $this->convertToPx($v);
                        break;
                    case '-syno-alt':
                        $arr['synoAlt'] = trim($v);
                        break;
                    case '-syno-file-id':
                        $arr['synoFileId'] = trim($v);
                        break;
                    case '-syno-file-type':
                        $arr['synoFileType'] = trim($v);
                        break;
                }
            }
        }
        return $arr;
    }

    private function convertToPx($value)
    {
        preg_match('/(\d+)(.*)/', $value, $match);
        $val = intval($match[1]);
        $unit = $match[2];
        switch($unit) {
            case 'in':
                $val = \PhpOffice\PhpWord\Shared\Converter::inchToPixel($val);
                break;
            case 'pt':
                $val = \PhpOffice\PhpWord\Shared\Converter::pointToPixel($val);
                break;
            default:
                break;
        }
        return $val;
    }

    /**
     * Read w:tblPr
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @return string|array|null
     * @todo Capture w:tblStylePr w:type="firstRow"
     */
    protected function readTableStyle(XMLReader $xmlReader, \DOMElement $domNode)
    {
        if (!$xmlReader->elementExists('w:tblPr', $domNode)) {
            return null;
        }
        $styleNode = $xmlReader->getElement('w:tblPr', $domNode);
        $styleDefs = array(
            'styleName' => array(self::READ_VALUE, 'w:tblStyle'));
        $margins = array('top', 'left', 'bottom', 'right');
        foreach ($margins as $side) {
            $ucfSide = ucfirst($side);
            $styleDefs["cellMargin$ucfSide"] = array(self::READ_VALUE, "w:tblCellMar/w:$side", 'w:w');
        }
        $borders = array_merge($margins, array('insideH', 'insideV'));
        foreach ($borders as $side) {
            $ucfSide = ucfirst($side);
            $styleDefs["border{$ucfSide}Size"] = array(self::READ_VALUE, "w:tblBorders/w:$side", 'w:sz');
            $styleDefs["border{$ucfSide}Color"] = array(self::READ_VALUE, "w:tblBorders/w:$side", 'w:color');
            $styleDefs["border{$ucfSide}Style"] = array(self::READ_VALUE, "w:tblBorders/w:$side", 'w:val');
        }
        $styleDefs['align'] = array(self::READ_VALUE, 'w:jc');
        return $this->readStyleDefs($xmlReader, $styleNode, $styleDefs);
    }

    /**
     * Read w:tcPr
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @return array
     */
    private function readCellStyle(XMLReader $xmlReader, \DOMElement $domNode)
    {
        $styleDefs = array(
            'valign'        => array(self::READ_VALUE, 'w:vAlign'),
            'textDirection' => array(self::READ_VALUE, 'w:textDirection'),
            'gridSpan'      => array(self::READ_VALUE, 'w:gridSpan'),
            'vMerge'        => array(self::READ_VALUE, 'w:vMerge'),
            'bgColor'       => array(self::READ_VALUE, 'w:shd', 'w:fill'),
        );
        $margins = array('top', 'left', 'bottom', 'right');
        foreach ($margins as $side) {
            $ucfSide = ucfirst($side);
            $styleDefs["border{$ucfSide}Size"] = array(self::READ_VALUE, "w:tcBorders/w:$side", 'w:sz');
            $styleDefs["border{$ucfSide}Color"] = array(self::READ_VALUE, "w:tcBorders/w:$side", 'w:color');
            $styleDefs["border{$ucfSide}Style"] = array(self::READ_VALUE, "w:tcBorders/w:$side", 'w:val');
        }

        return $this->readStyleDefs($xmlReader, $domNode, $styleDefs);
    }

    /**
     * Read style definition
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $parentNode
     * @param array $styleDefs
     * @ignoreScrutinizerPatch
     * @return array
     */
    protected function readStyleDefs(XMLReader $xmlReader, \DOMElement $parentNode = null, $styleDefs = array())
    {
        $styles = array();

        foreach ($styleDefs as $styleProp => $styleVal) {
            @list($method, $element, $attribute, $expected) = $styleVal;

            if ($xmlReader->elementExists($element, $parentNode)) {
                $node = $xmlReader->getElement($element, $parentNode);

                // Use w:val as default if no attribute assigned
                $attribute = ($attribute === null) ? 'w:val' : $attribute;
                $attributeValue = $xmlReader->getAttribute($attribute, $node);

                $styleValue = $this->readStyleDef($method, $attributeValue, $expected);

                // The w:val of w:vMerge should be either 'restart' or 'continue'
                // But in some documents, the continued w:vMerge have no w:val
                // Assign it 'continue' by default
                if (($element === 'w:vMerge') && !$styleValue) {
                    $styleValue = 'continue';
                }

                if ($styleValue !== null) {
                    $styles[$styleProp] = $styleValue;
                }
            }
        }

        return $styles;
    }

    /**
     * Return style definition based on conversion method
     *
     * @param string $method
     * @ignoreScrutinizerPatch
     * @param mixed $attributeValue
     * @param mixed $expected
     * @return mixed
     */
    private function readStyleDef($method, $attributeValue, $expected)
    {
        $style = $attributeValue;

        if ($method == self::READ_SIZE) {
            $style = $attributeValue / 2;
        } elseif ($method == self::READ_TRUE) {
            if ($attributeValue === '0' || $attributeValue === 'false') {
                $style = false;
            } else {
                $style = true;
            }
        } elseif ($method == self::READ_FALSE) {
            $style = false;
        } elseif ($method == self::READ_EQUAL) {
            $style = $attributeValue == $expected;
        }

        return $style;
    }

    /**
     * Returns the target of image, object, or link as stored in ::readMainRels
     *
     * @param string $docPart
     * @param string $rId
     * @return string|null
     */
    private function getMediaTarget($docPart, $rId)
    {
        $target = null;

        if (isset($this->rels[$docPart]) && isset($this->rels[$docPart][$rId])) {
            $target = $this->rels[$docPart][$rId]['target'];
        }

        return $target;
    }

    /**
     * Read w:sdt.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLReader $xmlReader
     * @param \DOMElement $domNode
     * @param mixed $parent
     * @param string $docPart
     * @return void
     *
     */
    protected function readTOC(XMLReader $xmlReader, \DOMElement $domNode, $parent, $docPart = 'document')
    {
        $nodes = $xmlReader->getElements('w:p', $domNode);
        foreach ($nodes as $node) {
            $this->readParagraph($xmlReader, $node, $parent);
        }
    }

    /**
     * Read w:drawing.
     */
    protected function readDrawing(XMLReader $xmlReader, \DOMElement $domNode, $parent, $docPart, $paragraphStyle = null, $fontStyle = null)
    {
        $readPositioning = function ($xmlReader, &$domNode) {
            if ($xmlReader->elementExists('wp:inline', $domNode)) {
                // TODO
                $domNode = $xmlReader->getElement('wp:inline', $domNode);
                return null;
            } elseif ($xmlReader->elementExists('wp:anchor', $domNode)) {
                // TODO
                $domNode = $xmlReader->getElement('wp:anchor', $domNode);
                return null;
            } else {
                throw new \Exception("readPositioning failed: no wp:inline or wp:anchor in {$domNode->nodeName}");
            }
        };

        $readProperty = function ($xmlReader, $domNode) {
            if (!$xmlReader->elementExists('wp:docPr', $domNode)) {
                return array(
                    'id'    => null,
                    'name'  => null,
                    'title' => null,
                    'desc'  => null);
            }

            $domNode = $xmlReader->getElement('wp:docPr', $domNode);
            return array(
                'id'    => $xmlReader->getAttribute('id', $domNode),
                'name'  => $xmlReader->getAttribute('name', $domNode),
                'title' => $xmlReader->getAttribute('title', $domNode),
                'desc'  => $xmlReader->getAttribute('desc', $domNode));
        };

        $readDimension = function ($xmlReader, $domNode) {
            if (!$xmlReader->elementExists('wp:extent', $domNode)) {
                return array(
                    'width'  => null,
                    'height' => null);
            }

            $domNode = $xmlReader->getElement('wp:extent', $domNode);
            return array(
                'width'  => Drawing::emuToPixels($xmlReader->getAttribute('cx', $domNode)),
                'height' => Drawing::emuToPixels($xmlReader->getAttribute('cy', $domNode)));
        };

        $readGraphic = function ($xmlReader, $domNode) {
            $readPic = function ($xmlReader, $domNode) {
                $readSource = function ($xmlReader, $domNode) {
                    if (!$xmlReader->elementExists('pic:blipFill/a:blip', $domNode)) {
                        return null;
                    }
                    return $xmlReader->getAttribute('r:embed', $xmlReader->getElement('pic:blipFill/a:blip', $domNode));
                };

                $readCrop = function ($xmlReader, $domNode) {
                    if (!$xmlReader->elementExists('pic:blipFill/a:srcRect', $domNode)) {
                        return null;
                    }
                    $srcRect = $xmlReader->getElement('pic:blipFill/a:srcRect', $domNode);
                    $b = intval($xmlReader->getAttribute('b', $srcRect));
                    $l = intval($xmlReader->getAttribute('l', $srcRect));
                    $r = intval($xmlReader->getAttribute('r', $srcRect));
                    $t = intval($xmlReader->getAttribute('t', $srcRect));
                    return array(
                        'b' => $b,
                        'l' => $l,
                        'r' => $r,
                        't' => $t
                    );
                };

                if (!$xmlReader->elementExists('pic:pic', $domNode)) {
                    // .docx files exported by Microsoft Word use namespace
                    // xmlns:pic="http://schemas.openxmlformats.org/drawingml/2006/picture"
                    $path = $xmlReader->getPath();
                    $path->registerNamespace('pic', 'http://schemas.openxmlformats.org/drawingml/2006/picture');
                    if (!$xmlReader->elementExists('pic:pic', $domNode)) {
                        throw new \Exception("readPic failed: no pic:pic in {$domNode->nodeName}");
                    }
                }

                $domNode = $xmlReader->getElement('pic:pic', $domNode);
                return array(
                    'type' => 'image',
                    'src'  => $readSource($xmlReader, $domNode),
                    'crop' => $readCrop($xmlReader, $domNode));
            };

            $readChart = function ($xmlReader, $domNode) {
                if (!$xmlReader->elementExists('c:chart', $domNode)) {
                    // .docx files exported by Microsoft Word use namespace
                    // xmlns:c="http://schemas.openxmlformats.org/drawingml/2006/chart"
                    $path = $xmlReader->getPath();
                    $path->registerNamespace('c', 'http://schemas.openxmlformats.org/drawingml/2006/chart');
                    if (!$xmlReader->elementExists('c:chart', $domNode)) {
                        throw new \Exception("readChart failed: no c:chart in {$domNode->nodeName}");
                    }
                }

                $domNode = $xmlReader->getElement('c:chart', $domNode);
                $rId = $xmlReader->getAttribute('r:id', $domNode);
                return array(
                    'type' => 'chart',
                    'rid'  => $rId
                );
            };

            $readGraph = function ($xmlReader, $domNode) {
                $readSolidFill = function ($xmlReader, $domNode, &$config, $configName) {
                    $solidFill = $xmlReader->getElement('a:solidFill', $domNode);
                    if (!$solidFill) {
                        return;
                    }
                    $color = $xmlReader->getElement('a:srgbClr', $solidFill); // TODO other color tags
                    if (!$color) {
                        return;
                    }
                    $config[$configName] = "#{$xmlReader->getAttribute('val', $color)}";
                };

                $readLineEnd = function ($xmlReader, $domNode, $configName, &$config) {
                    if (!$domNode) {
                        return;
                    }
                    switch ($xmlReader->getAttribute('type', $domNode)) {
                        case 'triangle':
                            $config[$configName] = 'arrow';
                            break;
                        default:
                            break;
                    }
                };

                if (!$xmlReader->elementExists('wps:wsp', $domNode)) {
                    $path = $xmlReader->getPath();
                    $path->registerNamespace('wps', 'http://schemas.microsoft.com/office/word/2010/wordprocessingShape');
                    if (!$xmlReader->elementExists('wps:wsp', $domNode)) {
                        throw new \Exception("readChart failed: no wps:wsp in {$domNode->nodeName}");
                    }
                }

                $config = array(
                    'alignment' => 'inline', // TODO
                    'customized_points' => array(),
                    'order' => 1, // TODO
                );
                $domNode = $xmlReader->getElement('wps:wsp', $domNode);
                $spPr = $xmlReader->getElement('wps:spPr', $domNode);
                if ($spPr) {
                    $xfrm = $xmlReader->getElement('a:xfrm', $spPr);
                    if ($xfrm) {
                        $config['flip_h'] = $xmlReader->getAttribute('flipH', $xfrm) === '1';
                        $config['flip_v'] = $xmlReader->getAttribute('flipV', $xfrm) === '1';
                        $config['deg'] = intval($xmlReader->getAttribute('rot', $xfrm)) / 60000;
                        $ext = $xmlReader->getElement('a:ext', $xfrm);
                        if ($ext) {
                            $config['width'] = intval($xmlReader->getAttribute('cx', $ext)) / 9525;
                            $config['height'] = intval($xmlReader->getAttribute('cy', $ext)) / 9525;
                        }
                    }
                    $prstGeom = $xmlReader->getElement('a:prstGeom', $spPr);
                    if ($prstGeom) {
                        $config['prst'] = $xmlReader->getAttribute('prst', $prstGeom);
                        $avLst = $xmlReader->getElements('a:avLst/a:gd', $prstGeom);
                        foreach ($avLst as $gd) {
                            $name = $xmlReader->getAttribute('name', $gd);
                            $fmla = $xmlReader->getAttribute('fmla', $gd);
                            if (!($name && $fmla)) {
                                continue;
                            }
                            $config['customized_points'][$name] = preg_split("/\s+/", trim($fmla))[1]; // assuming fmla must be in the format "val <value>"
                        }
                    }
                    $readSolidFill($xmlReader, $spPr, $config, 'fill');
                    $ln = $xmlReader->getElement('a:ln', $spPr);
                    if ($ln) {
                        $w = $xmlReader->getAttribute('w', $ln);
                        if ($w) {
                            $config['stroke-width'] = intval($w) / 9525;
                        }
                        $readSolidFill($xmlReader, $ln, $config, 'stroke');
                        $readLineEnd($xmlReader, $xmlReader->getElement('a:headEnd', $ln), 'start_marker', $config);
                        $readLineEnd($xmlReader, $xmlReader->getElement('a:tailEnd', $ln), 'end_marker', $config);
                    }
                }
                return array(
                    'type' => 'graph',
                    'config' => $config,
                );
            };

            if (!$xmlReader->elementExists('a:graphic/a:graphicData', $domNode)) {
                // .docx files exported by Microsoft Word use namespace
                // xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main"
                $path = $xmlReader->getPath();
                $path->registerNamespace('a', 'http://schemas.openxmlformats.org/drawingml/2006/main');
                if (!$xmlReader->elementExists('a:graphic/a:graphicData', $domNode)) {
                    throw new \Exception("readGraphic failed: no a:graphic/a:graphicData in {$domNode->nodeName}");
                }
            }

            $domNode = $xmlReader->getElement('a:graphic/a:graphicData', $domNode);
            $uri = $xmlReader->getAttribute('uri', $domNode);
            switch ($uri) {
                case 'http://schemas.openxmlformats.org/drawingml/2006/picture': {
                    return $readPic($xmlReader, $domNode);
                }
                case 'http://schemas.openxmlformats.org/drawingml/2006/chart': {
                    return $readChart($xmlReader, $domNode);
                }
                case 'http://schemas.microsoft.com/office/word/2010/wordprocessingShape': {
                    return $readGraph($xmlReader, $domNode);
                }
                default: {
                    throw new \Exception("readGraphic failed: unknown uri {$uri}");
                }
            }
        };

        try {
            $positioning = $readPositioning($xmlReader, $domNode);
            $property = $readProperty($xmlReader, $domNode);
            $dimension = $readDimension($xmlReader, $domNode);
            $graphic = $readGraphic($xmlReader, $domNode);
            switch ($graphic['type']) {
                case 'image': {
                    if ($graphic['src'] === null) {
                        throw new \Exception('readDrawing failed: image source is not defined');
                    }

                    $target = $this->getMediaTarget($docPart, $graphic['src']);
                    if ($target === null) {
                        throw new \Exception("getMediaTarget failed, docPart:{$docPart} rId:{$graphic['src']}");
                    }
                    $src = "zip://{$this->docFile}#{$target}";
                    $style = array(
                        'width'        => $dimension['width'],
                        'height'       => $dimension['height'],
                        'synoFileType' => 'image',
                        'synoFileId'   => $property['id'],
                        'synoAlt'      => $property['name'],
                        'synoCrop'     => $graphic['crop']
                    );
                    $image = $parent->addImage($src, $style);
                    $image->setAnnotations($fontStyle);
                    break;
                }
                case 'chart': {
                    if ($graphic['rid'] === null) {
                        throw new \Exception('readDrawing failed: chart rid is not defined');
                    }

                    $target = $this->getMediaTarget($docPart, $graphic['rid']);
                    if ($target === null) {
                        throw new \Exception("getMediaTarget failed, docPart:{$docPart} rId:{$graphic['rid']}");
                    }

                    $xmlReader = new XMLReader();
                    $xmlReader->getDomFromZip($this->docFile, $target);
                    $config = $this->parseChartConfig($xmlReader);
                    try {
                        $chartType = $config["plotArea"]["subCharts"][0]["type"];
                    } catch (\Exception $e) {
                        throw new \Exception('readDrawing failed: chart type not defined');
                    }
                    $chart = $parent->addChart($chartType, array(), array());
                    $chart->setSynoChartConfig($config);
                    if (empty($dimension['width']) || empty($dimension['height'])) {
                        $dimension['width'] = 520;
                        $dimension['height'] = 380;
                    }
                    $chart->getStyle()->setWidth(\PhpOffice\PhpWord\Shared\Converter::pixelToEmu($dimension['width']));
                    $chart->getStyle()->setHeight(\PhpOffice\PhpWord\Shared\Converter::pixelToEmu($dimension['height']));
                    $chart->setAnnotations($fontStyle);
                    break;
                }
                case 'graph': {
                    $graph = $parent->addGraph($graphic['config']);
                    $graph->setAnnotations($fontStyle);
                    break;
                }
                default: {
                    throw new \Exception("readDrawing failed: unknown graphic type {$graphic['type']}");
                }
            }
        } catch (\Exception $e) {
            print_log($e->getMessage());
        }
    }

    private function parseChartConfig(XMLReader $xmlReader)
    {
        $titleNode = $xmlReader->getElement('c:chart/c:title');
        $config = array();
        $config["title"] = $this->readTitleBlock($xmlReader, $titleNode);
        $config["cornerData"] = "";
        $config["plotArea"] = array();
        $config["plotArea"]["xAxisTitle"] = "";
        $config["plotArea"]["yAxisTitle"] = "";
        $config["plotArea"]["subCharts"] = array();

        $plotNode = $xmlReader->getElement('c:chart/c:plotArea');
        foreach ($plotNode->childNodes as $node) {
            $item = array();
            switch ($node->nodeName) {
            case 'c:barChart':
                $barDirMapper = array("col" => "col", "bar" => "row");
                $item["type"] = "bar";
                $item["series"] = $this->parseChartSeries($xmlReader, $node);
                $item["barDir"] = $barDirMapper[$xmlReader->getAttribute('val', $node, 'c:barDir')];
                $item["grouping"] = $xmlReader->getAttribute('val', $node, 'c:grouping');
                $config["plotArea"]["subCharts"][] = $item;
                break;
            case 'c:lineChart':
                $item["type"] = "line";
                $item["series"] = $this->parseChartSeries($xmlReader, $node);
                $item["grouping"] = $xmlReader->getAttribute('val', $node, 'c:grouping');
                $config["plotArea"]["subCharts"][] = $item;
                break;
            case 'c:areaChart':
                $item["type"] = "area";
                $item["series"] = $this->parseChartSeries($xmlReader, $node);
                $item["grouping"] = $xmlReader->getAttribute('val', $node, 'c:grouping');
                $config["plotArea"]["subCharts"][] = $item;
                break;
            case 'c:pieChart':
                $item["type"] = "pie";
                $item["series"] = $this->parseChartSeries($xmlReader, $node);
                $item["grouping"] = $xmlReader->getAttribute('val', $node, 'c:grouping');
                $config["plotArea"]["subCharts"][] = $item;
                break;
            case 'c:scatterChart':
                $item["type"] = "scatter";
                $item["series"] = $this->parseChartSeries($xmlReader, $node);
                $item["grouping"] = $xmlReader->getAttribute('val', $node, 'c:grouping');
                $config["plotArea"]["subCharts"][] = $item;
                break;
            case 'c:catAx':
                $title = $xmlReader->getElement('c:title', $node);
                $config["plotArea"]["xAxisTitle"] = $this->readTitleBlock($xmlReader, $title);
                break;
            case 'c:valAx':
                $title = $xmlReader->getElement('c:title', $node);
                $config["plotArea"]["yAxisTitle"] = $this->readTitleBlock($xmlReader, $title);
                break;
            default:
                break;
            }
        }
        return $config;
    }

    private function readTitleBlock($xmlReader, $node)
    {
        $titleNode = $xmlReader->getElement('c:tx/c:rich/a:p/a:r', $node);
        $title = $xmlReader->getValue('a:t', $titleNode);
        return empty($title) ? "" : $title;
    }

    private function parseChartSeries($xmlReader, $node)
    {
        $nodes = $xmlReader->getElements('c:ser', $node);
        $series = array();
        foreach ($nodes as $node) {
            $item = array();
            $item["index"] = intval($xmlReader->getAttribute('val', $node, 'c:idx'));
            $item["order"] = intval($xmlReader->getAttribute('val', $node, 'c:order'));
            $marker = $xmlReader->getAttribute('val', $node, 'c:marker/c:symbol');
            $item["marker"] = !empty($marker) && $marker !== 'none';
            $item["tx"] = $xmlReader->getValue('c:tx/c:strRef/c:strCache/c:pt/c:v', $node);

            // parse category
            $item["cat"] = array();
            $catNode = $xmlReader->getElement('c:cat', $node);
            $isValue = false;
            if (empty($catNode)) {
                $catNode = $xmlReader->getElement('c:xVal', $node);
                $isValue = true;
            }
            $catList = $xmlReader->getElements('c:strLit/c:pt|c:strRef/c:strCache/c:pt|c:numRef/c:numCache/c:pt', $catNode);
            foreach ($catList as $row) {
                if ($isValue) {
                    $item["cat"][] = floatval($xmlReader->getValue('c:v', $row));
                } else {
                    $item["cat"][] = $xmlReader->getValue('c:v', $row);
                }
            }

            // parse value
            $item["val"] = array();
            $valNode = $xmlReader->getElement('c:val|c:yVal', $node);
            $valList = $xmlReader->getElements('c:numLit/c:pt|c:numRef/c:numCache/c:pt|c:numRef/c:numCache/c:pt', $valNode);
            foreach ($valList as $row) {
                $item["val"][] = floatval($xmlReader->getValue('c:v', $row));
            }

            $series[] = $item;
        }
        return $series;
    }
}
