BEGIN;

CREATE TABLE IF NOT EXISTS db_version (
	version NUMERIC NOT NULL
);

CREATE UNIQUE INDEX db_version_one_row ON db_version((version IS NOT NULL));

CREATE FUNCTION db_version_no_delete ()
RETURNS trigger
LANGUAGE plpgsql AS $f$
BEGIN
	RAISE EXCEPTION 'You may not delete the DB version!';
END; $f$;

CREATE TRIGGER db_version_no_delete
BEFORE DELETE ON db_version
FOR EACH ROW EXECUTE PROCEDURE db_version_no_delete();

INSERT INTO db_version (version) VALUES (9);

CREATE TABLE IF NOT EXISTS notebook (
	object_id TEXT UNIQUE NOT NULL,
	category TEXT DEFAULT 'notebook' NOT NULL,
	title TEXT NOT NULL,
	preset BOOLEAN DEFAULT FALSE,
	acl JSON,
	owner BIGINT NOT NULL,
	version TEXT NOT NULL,
	ctime BIGINT,
	mtime BIGINT,
	link_id TEXT NOT NULL,
	individual_shared BOOLEAN DEFAULT FALSE,
	stack TEXT NOT NULL,
	archive BOOL NOT NULL DEFAULT FALSE,
	extra_info JSON
);

CREATE INDEX idx_notebook_object_id ON notebook(object_id);
CREATE INDEX idx_notebook_ctime ON notebook(ctime);
CREATE INDEX idx_notebook_mtime ON notebook(mtime);

CREATE TABLE IF NOT EXISTS note (
	object_id TEXT UNIQUE NOT NULL,
	category TEXT DEFAULT 'note' NOT NULL,
	commit_msg JSON,
	latitude REAL,
	longitude REAL,
	location TEXT,
	title TEXT NOT NULL,
	brief TEXT,
	acl JSON,
	owner BIGINT NOT NULL,
	version TEXT NOT NULL,
	ctime BIGINT,
	mtime BIGINT,
	parent_id TEXT NOT NULL,
	encrypt BOOLEAN DEFAULT FALSE,
	recycle BOOLEAN DEFAULT FALSE,
	thumb TEXT,
	tag JSON,
	attachment JSON,
	old_parent_id TEXT,
	link_id TEXT NOT NULL,
	individual_shared BOOLEAN DEFAULT FALSE,
	archive BOOLEAN DEFAULT FALSE,
	extra_info JSON
);

CREATE INDEX idx_note_object_id ON note(object_id);
CREATE INDEX idx_note_title ON note(title);
CREATE INDEX idx_note_ctime ON note(ctime);
CREATE INDEX idx_note_mtime ON note(mtime);
CREATE INDEX idx_note_parent_id ON note(parent_id);
CREATE INDEX idx_note_recycle_and_owner ON note(recycle, owner);

CREATE TABLE IF NOT EXISTS note_encrypt (
	object_id TEXT NOT NULL,
	ctime BIGINT NOT NULL,
	mtime BIGINT NOT NULL,
	duration INTEGER NOT NULL,
	token TEXT UNIQUE NOT NULL,
	password TEXT NOT NULL
);

CREATE INDEX idx_note_encrypt_object_id ON note_encrypt(object_id);

CREATE TABLE IF NOT EXISTS link (
	id TEXT UNIQUE NOT NULL,
	object_id TEXT NOT NULL,
	category TEXT NOT NULL
);

CREATE INDEX idx_link_id ON link(id);
CREATE INDEX idx_link_object_id ON link(object_id);

CREATE TABLE IF NOT EXISTS perm_public (
	object_id TEXT NOT NULL,
	perm INTEGER NOT NULL,
	inherit BOOLEAN NOT NULL
);

CREATE INDEX idx_perm_public_object_id ON perm_public(object_id);

CREATE TABLE IF NOT EXISTS perm_user (
	object_id TEXT NOT NULL,
	uid BIGINT NOT NULL,
	perm INTEGER NOT NULL,
	inherit BOOLEAN NOT NULL
);

CREATE INDEX idx_perm_user_object_id ON perm_user(object_id);

CREATE TABLE IF NOT EXISTS perm_group (
	object_id TEXT NOT NULL,
	gid BIGINT NOT NULL,
	perm INTEGER NOT NULL,
	inherit BOOLEAN NOT NULL
);

CREATE INDEX idx_perm_group_object_id ON perm_group(object_id);

CREATE TABLE IF NOT EXISTS tag_note (
	note_object_id TEXT NOT NULL,
	tag_id TEXT NOT NULL
);

CREATE TABLE IF NOT EXISTS tag (
	id TEXT UNIQUE NOT NULL,
	title TEXT NOT NULL,
	category TEXT DEFAULT 'tag' NOT NULL,
	owner BIGINT NOT NULL,
	extra_info JSON
);

CREATE INDEX idx_tag_title ON tag(title);

CREATE TABLE IF NOT EXISTS object_shortcut (
	owner BIGINT NOT NULL,
	object_id TEXT NOT NULL
);

CREATE INDEX idx_object_shortcut_object_id ON object_shortcut(object_id);

CREATE TABLE IF NOT EXISTS tag_shortcut (
	owner BIGINT NOT NULL,
	tag_id TEXT NOT NULL
);

CREATE INDEX idx_tag_shortcut_tag_id ON tag_shortcut(tag_id);

CREATE TABLE IF NOT EXISTS stack (
	name TEXT NOT NULL,
	owner BIGINT NOT NULL,
	notebook_id TEXT UNIQUE NOT NULL
);

CREATE INDEX idx_stack_name_and_owner ON stack(name, owner);

CREATE TABLE IF NOT EXISTS stack_shortcut (
	owner BIGINT NOT NULL,
	stack_name TEXT NOT NULL
);

CREATE INDEX idx_stack_shortcut_owner_and_stack_name ON stack_shortcut(stack_name, owner);

CREATE TABLE IF NOT EXISTS todo (
	object_id TEXT UNIQUE  NOT NULL,
	owner BIGINT NOT NULL,
	note_id TEXT NOT NULL,
	title TEXT NOT NULL,
	ctime BIGINT NOT NULL DEFAULT 0,
	mtime BIGINT NOT NULL DEFAULT 0,
	due_date BIGINT,
	reminder_offset BIGINT,
	priority INTEGER,
	encrypt BOOLEAN DEFAULT FALSE,
	done BOOLEAN DEFAULT FALSE,
	complete BIGINT NOT NULL DEFAULT 0,
	parent_id TEXT,
	star BOOLEAN DEFAULT FALSE,
	comment TEXT DEFAULT ''
);

CREATE TABLE IF NOT EXISTS smart (
	object_id TEXT UNIQUE NOT NULL,
	link_id TEXT NOT NULL,
	category TEXT DEFAULT 'smart' NOT NULL,
	title TEXT NOT NULL,
	acl JSON,
	owner BIGINT NOT NULL,
	ctime BIGINT,
	mtime BIGINT,
	individual_shared BOOLEAN DEFAULT FALSE,
	query JSON
);

CREATE INDEX idx_smart_object_id ON smart(object_id);
CREATE INDEX idx_smart_ctime ON smart(ctime);
CREATE INDEX idx_smart_mtime ON smart(mtime);

CREATE OR REPLACE FUNCTION afterNotebookDeleted()
RETURNS "trigger" AS $$
	BEGIN
		DELETE FROM perm_public WHERE object_id=OLD.object_id;
		DELETE FROM perm_user WHERE object_id=OLD.object_id;
		DELETE FROM perm_group WHERE object_id=OLD.object_id;
		DELETE FROM object_shortcut WHERE object_id=OLD.object_id;
		DELETE FROM link WHERE object_id=OLD.object_id;
		DELETE FROM stack WHERE notebook_id=OLD.object_id;
		RETURN NULL;
	END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION afterNoteDeleted()
RETURNS "trigger" AS $$
	BEGIN
		DELETE FROM tag_note WHERE note_object_id=OLD.object_id;
		DELETE FROM perm_public WHERE object_id=OLD.object_id;
		DELETE FROM perm_user WHERE object_id=OLD.object_id;
		DELETE FROM perm_group WHERE object_id=OLD.object_id;
		DELETE FROM object_shortcut WHERE object_id=OLD.object_id;
		DELETE FROM link WHERE object_id=OLD.object_id;
		DELETE FROM note_encrypt WHERE object_id=OLD.object_id;
		RETURN NULL;
	END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION afterTagRelationDeleted()
RETURNS "trigger" AS $$
	BEGIN
		DELETE FROM tag
		WHERE 0=(SELECT count(tag_id) FROM tag_note WHERE tag_id=OLD.tag_id)
		AND id=OLD.tag_id;
		RETURN NULL;
	END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION removeStackFromShortcut(name TEXT, owner BIGINT)
RETURNS boolean AS '/var/packages/NoteStation/target/trigger/pgtrigger.so', 'stack_deleted'
LANGUAGE C SECURITY DEFINER;

CREATE OR REPLACE FUNCTION afterStackDeleted()
RETURNS "trigger" AS $$
	BEGIN
		IF 0 = (SELECT count(name) FROM stack WHERE name=OLD.name AND owner=OLD.owner) THEN
			PERFORM removeStackFromShortcut(OLD.name, OLD.owner);
		END IF;
		RETURN NULL;
	END;
$$ LANGUAGE plpgsql;

CREATE OR REPLACE FUNCTION afterTagDeleted()
RETURNS "trigger" AS '/var/packages/NoteStation/target/trigger/pgtrigger.so', 'tag_deleted'
LANGUAGE C SECURITY DEFINER;

CREATE OR REPLACE FUNCTION afterSmartDeleted()
RETURNS "trigger" AS $$
	BEGIN
		DELETE FROM perm_public WHERE object_id=OLD.object_id;
		DELETE FROM perm_user WHERE object_id=OLD.object_id;
		DELETE FROM perm_group WHERE object_id=OLD.object_id;
		DELETE FROM object_shortcut WHERE object_id=OLD.object_id;
		DELETE FROM link WHERE object_id=OLD.object_id;
		RETURN NULL;
	END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER "onDelete" AFTER DELETE
   ON notebook
   FOR EACH ROW
   EXECUTE PROCEDURE afterNotebookDeleted();

CREATE TRIGGER "onDelete" AFTER DELETE
   ON note
   FOR EACH ROW
   EXECUTE PROCEDURE afterNoteDeleted();

CREATE TRIGGER "onDelete" AFTER DELETE
   ON tag_note
   FOR EACH ROW
   EXECUTE PROCEDURE afterTagRelationDeleted();

CREATE TRIGGER "onDelete" AFTER DELETE
   ON tag
   FOR EACH ROW
   EXECUTE PROCEDURE afterTagDeleted();

CREATE TRIGGER "onDelete" AFTER DELETE
   ON stack
   FOR EACH ROW
   EXECUTE PROCEDURE afterStackDeleted();

CREATE TRIGGER "onDelete" AFTER DELETE
   ON smart
   FOR EACH ROW
   EXECUTE PROCEDURE afterSmartDeleted();

COMMIT;
